/* ***** BEGIN LICENSE BLOCK *****
Version: MPL 1.1/GPL 2.0/LGPL 2.1

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is Standard-Sitemap Navigator code.

The Initial Developer of the Original Code is the
Standard-Sitemap Team.

Contributor(s): (see license.txt)

Alternatively, the contents of this file may be used under the terms of
either the GNU General Public License Version 2 or later (the "GPL"), or
the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
in which case the provisions of the GPL or the LGPL are applicable instead
of those above. If you wish to allow use of your version of this file only
under the terms of either the GPL or the LGPL, and not to allow others to
use your version of this file under the terms of the MPL, indicate your
decision by deleting the provisions above and replace them with the notice
and other provisions required by the GPL or the LGPL. If you do not delete
the provisions above, a recipient may use your version of this file under
the terms of any one of the MPL, the GPL or the LGPL.
***** END LICENSE BLOCK ***** */

/* $Id: sslib.js 484 2011-02-08 00:27:29Z ssjools $
 *
 * Generic utility functions. Any function that has nothing to do with Standard Sitemap
 * and could be used in any extension should be stored here.
 */


if (!Application.storage.get("org.standard-sitemap.library", null)) {
    Application.storage.set("org.standard-sitemap.library", {
        Components : Components,

        extId      : 'extension@standard-sitemap.org',
        extName    : 'org.standard-sitemap.navigator',

        cout       : Components.classes["@mozilla.org/consoleservice;1"]
            .getService(Components.interfaces.nsIConsoleService),

        ioService  : Components.classes["@mozilla.org/network/io-service;1"]
            .getService(Components.interfaces.nsIIOService),

        oObserverService :
        Components.classes["@mozilla.org/observer-service;1"]
            .getService(Components.interfaces.nsIObserverService),

        initialize : function() {
            var sbService =
                Components.classes["@mozilla.org/intl/stringbundle;1"]
                .getService(Components.interfaces.nsIStringBundleService);
            this.oStringBundle = sbService.createBundle
            ("chrome://standardsitemap/locale/standardsitemap.properties");
            this.oLanguageBundle = sbService.createBundle
            ("chrome://global/locale/languageNames.properties");
            this.oRegionBundle = sbService.createBundle
            ("chrome://global/locale/regionNames.properties");

            /* Get the current browsing mode. */
            // https://developer.mozilla.org/En/Supporting_private_browsing_mode
            try {
                var pbs = Components.classes["@mozilla.org/privatebrowsing;1"]
                    .getService(Components.interfaces
                                .nsIPrivateBrowsingService);
                this.bIsPrivate = pbs.privateBrowsingEnabled;
            } catch(ex) {
                // Ignore exceptions in older versions of Firefox.
            }

            // Set up an observer to watch for changes in browsing
            // mode.
            var self = this;
            var privateWatcher = {
                observe : function(aSubject, aTopic, aData) {
                    if (aTopic == "quit-application") {
                        self.oObserverService
                            .removeObserver(self, "quit-application");
                        self.oObserverService
                            .removeObserver(self, "private-browsing");
                        return;
                    }
                    if (aTopic == "private-browsing") {
                        if (aData == "enter")
                            self.enterPrivateBrowsing();
                        else if (aData == "exit")
                            self.exitPrivateBrowsing();
                        return;
                    }
                },
            };
            this.oObserverService
                .addObserver(privateWatcher, "private-browsing", false);
            this.oObserverService
                .addObserver(privateWatcher, "quit-application", false);

            this.initPrefs();
        },

        dump : function(msg) {
            dump('[STDMAP] ' + msg + '\n');
            this.debug(msg);
        },
    
        debug : function(str) {
            if (this.prefs.debug) {
                this.cout.logStringMessage(this.stringSubst
                                           ("%1 (%3): %2", this.extName,
                                            str, new Date().getTime()));
            }
        },

        listMembers : function(obj) {
            if (!this.prefs.debug) return;
            var s = 'Members of ' + obj + ':\n';
            for (var key in obj) {
                s += '\t' + key + '\n';
            }
            this.debug(s);
        },

        makeURI : function(text) {
            if (text == null) return null;
            return this.ioService.newURI(text, 'us-ascii', null);
        },

        // Get xml:base of a node by looking at its ancestry for
        // xml:base.  'node' should be a DOM node.
        getBaseURI: function(oNode) {
            // TODO: When we drop support for @baseurl, the
            // implementation reduces to this single line.  Until
            // then, it is disabled, as it won't recognise @baseurl as
            // a back-up to @xml:base.

            //return this.makeURI(oNode.baseURI);

            // The following implementation uses @baseurl as a back-up
            // to @xml:base.

            // Have we worked this out before?
            if (oNode.sslib_base)
                return oNode.sslib_base;

            // Get the context for any URI specified at this node.
            var oContext = null;
            if ((oNode.parentNode &&
                 (oContext = this.getBaseURI(oNode.parentNode),
                  true)) ||
                (oNode.ownerDocument &&
                 (oContext = this.makeURI(oNode.ownerDocument.documentURI),
                  true))) {
                // This is a document node.  If it's not the root
                // element, its context is its document's URI;
                // otherwise, it is its parent's URI.

                // Does this node define xml:base (or sitemap:baseurl)?
                var sRaw;
                if (oNode.ownerDocument &&
                    ((sRaw = oNode.getAttributeNS(this.xmlNS, 'base')) ||
                     (sRaw = this.getAttributeSM(oNode, 'baseurl')) ||
                     (this.issmns(oNode.namespaceURI) &&
                      (sRaw = oNode.getAttribute('baseurl'))))) {
            
                    // Yes, so extract it, resolve it against the
                    // context, and make an nsIURI object out of it.
                    oNode.sslib_base = this.makeURI(oContext.resolve(sRaw));
                } else {
                    // No, so the base of this element is the same as
                    // its parent.
                    oNode.sslib_base = oContext;
                }
            } else {
                // This is the whole document.  Just use its URI.
                oNode.sslib_base = this.makeURI(oNode.documentURI);
            }

            return oNode.sslib_base;
        },

        xmlNS : 'http://www.w3.org/XML/1998/namespace',

        xhtmlNS : 'http://www.w3.org/1999/xhtml',

        xlinkNS : 'http://www.w3.org/1999/xlink',

        sitemapNS : [
            'http://standard-sitemap.org/2007/ns',
            'http://www.standard-sitemap.org/version1beta/',
        ],

        xulNS : 'http://www.mozilla.org/keymaster/gatekeeper/there.is.only.xul',

        // Get an attribute in any of our namespaces.
        getAttributeSM : function(node, name) {
            for each (var cand in this.sitemapNS) {
                var val = node.getAttributeNS(cand, name);
                if (val)
                    return val;
            }
            return null;
        },

        // Get the elements of a particular tag name in any of our
        // namespaces.  The first to produce a non-empty list is
        // returned.
        getElementsByTagNameSM : function(node, name) {
            for each (var cand in this.sitemapNS) {
                var elems = node.getElementsByTagNameNS(cand, name);
                if (elems != null && elems.length > 0)
                    return elems;
            }
            return elems;
        },

        // Is this namespace one of ours?
        issmns : function(s) {
            var parts = s.trim().split('\\s+');
            for each (var part in parts) {
                for each (var cand in this.sitemapNS) {
                    if (cand == part)
                        return true;
                }
            }
            return false;
        },

        resolveInNode: function(oNode, sRaw) {
            if (!sRaw) return null
            var oBase = this.getBaseURI(oNode);
            return this.makeURI(oBase.resolve(sRaw));
        },

        getURIAttr: function(oNode, sName) {
            return this.resolveInNode(oNode, oNode.getAttribute(sName));
        },

        getURIAttrNS: function(oNode, sNamespace, sName) {
            return this.resolveInNode(oNode,
                                      oNode.getAttributeNS(sNamespace, sName));
        },

        objectToString : function(obj) {
            var s = '';
            var key;
            for (key in obj) {
                s += this.stringSubst("%1 : %2", key, obj[key]);
            }
            return s;
        },

        migratePrefs : function() {
            var names = [ 'modeWebmaster', 'referrer', 'debug',
                          'autoopen', 'urlbaricon', 'visitonselect',
                          'findonvisit', 'openonvisit', 'openonclick',
                          'selectonvisit', 'vertbar', 'showRoleItemsInTree',
                          'invertctrlforfocus', 'ctrlfornewtab',
                          'showUpButton', 'includeOptionalItems',
                          'installed' ];
            var oldBranch =
                Components.classes["@mozilla.org/preferences-service;1"]
                .getService(Components.interfaces.nsIPrefService)
                .getBranch("extensions.standardsitemap.");
            for each (var name in names) {
                // Only deal with old values that still exist.
                var oldType = oldBranch.getPrefType(name);
                if (oldType == 0) continue;
                var newType = this.prefBranch.getPrefType(name);
                switch (oldType) {
                case this.prefBranch.PREF_STRING:
                    this.prefBranch.setCharPref(name,
                                                oldBranch.getCharPref(name));
                    break;

                case this.prefBranch.PREF_INT:
                    this.prefBranch.setIntPref(name,
                                                oldBranch.getIntPref(name));
                    break;

                case this.prefBranch.PREF_BOOL:
                    this.prefBranch.setBoolPref(name,
                                                oldBranch.getBoolPref(name));
                    break;
                }
                oldBranch.clearUserPref(name);
            }
        },

        // Adds all prefs to a prefs object
        initPrefs : function() {
            this.prefBranch =
                Components.classes["@mozilla.org/preferences-service;1"]
                .getService(Components.interfaces.nsIPrefService)
                .getBranch(this.stringSubst("extensions.%1.", this.extName));
            this.prefBranch
                .QueryInterface(Components.interfaces.nsIPrefBranch2);
            this.prefBranch.addObserver("", this, false);

            //Make sure we don't leak
//             window.addEventListener('unload', function(event) {
//                 this.unregister();
//             }, false);
        
            this.prefs = {};

            this.migratePrefs();

            var list = this.prefBranch.getChildList("", {})
                .toString().split(",");

            for (var i in list) {
                var name = list[i];
                var type = this.prefBranch.getPrefType(name);

                if (type == this.prefBranch.PREF_STRING) {
                    this.prefs[name] = this.getCharPref(name);
                } else if (type == this.prefBranch.PREF_INT)  {
                    this.prefs[name] = this.getIntPref(name);
                } else if (type == this.prefBranch.PREF_BOOL) {
                    this.prefs[name] = this.getBoolPref(name);
                }
            }
            //this.setPref('autoopen', true); //set the specific pref
            //this.setPref('other', 'value'); //add new prefs
        },

        setPref : function(name, value) {
            var type = this.prefBranch.getPrefType(name);

            // undefined prefs, when adding new not specified in defaults
            if (type == 0) {
                if (typeof(value) == 'number') {
                    type = 64;
                } else if (typeof(value) == 'string') {
                    type = 32;
                } else if (typeof(value) == 'boolean') {
                    type = 128;
                }
            }

            if (type == this.prefBranch.PREF_STRING) {
                this.setCharPref(name, value);
            } else if (type == this.prefBranch.PREF_INT)  {
                this.setIntPref(name, value);
            } else if (type == this.prefBranch.PREF_BOOL) {
                this.setBoolPref(name, value);
            }
        },

//         unregister : function() {
//             if(!this.prefBranch) return;
//             this.prefBranch.removeObserver("", this);
//         },

        observe : function(aSubject, aTopic, aData) {
            if(aTopic != "nsPref:changed") return;
            // HELP it works! but sometimes get error SSLib.dump is
            // not defined, check on other plaforms.

            // Out of scope sometimes? Have not seen it for a while
            // now, caveman 03-02-2009
            if (typeof(this.dump) == "function") {
                this.dump("Pref " + aData + " changed.");
                this.oObserverService
                    .notifyObservers(null, "standard-sitemap-prefchanged",
                                     aData);
            }
        },

        getCharPref : function(branch) {
            return this.prefBranch.getCharPref(branch);
        },

        setCharPref : function(branch, value) {
            this.prefBranch.setCharPref(branch, value);
        },

        getBoolPref : function(branch) {
            return this.prefBranch.getBoolPref(branch);
        },

        setBoolPref : function(branch, value) {
            this.prefBranch.setBoolPref(branch, value);
        },

        getIntPref : function(branch) {
            return this.prefBranch.getIntPref(branch);
        },

        setIntPref : function(branch, value) {
            this.prefBranch.setIntPref(branch, value);
        },

        savedContentPrefs :
        Components.classes["@mozilla.org/content-pref/service;1"]
            .getService(Components.interfaces.nsIContentPrefService),

        privateContentPrefs : { },

        bIsPrivate : false,

        enterPrivateBrowsing : function() {
            this.bIsPrivate = true;
        },

        exitPrivateBrowsing : function() {
            this.bIsPrivate = false;
            this.privateContentPrefs = { };
        },

        getContentPref : function(uri, key) {
            if (this.bIsPrivate) {
                var pkey = this.getPrivateContentPrefKey(uri, key);
                var val = this.privateContentPrefs[pkey];
                if (val) return val;
            }
            uri = this.fixSitePref(uri);
            return this.savedContentPrefs.getPref(uri, key);
        },

        setContentPref : function(uri, key, value) {
            if (this.bIsPrivate) {
                var pkey = this.getPrivateContentPrefKey(uri, key);
                this.privateContentPrefs[pkey] = value;
            } else {
                uri = this.fixSitePref(uri);
                this.savedContentPrefs.setPref(uri, key, value);
            }
        },

        removeContentPref : function(uri, key) {
            if (this.bIsPrivate) {
                var pkey = this.getPrivateContentPrefKey(uri, key);
                delete this.privateContentPrefs[pkey];
            } else {
                uri = this.fixSitePref(uri);
                this.savedContentPrefs.removePref(uri, key);
            }
        },

        getPrivateContentPrefKey : function(uri, key) {
            if (!uri) return '>' + key;
            return uri.spec + '>' + key;
        },

        // The Content Preference Service uses a narrow definition of
        // 'site'.  In fact, it's handles host preferences, but we want
        // preferences per URI, not per host.  This function translates
        // one nsIURI into another, incorporating the path component into
        // a questionable domainname appended to the original host.
        fixSitePref : function(loc) {
            if (loc.asciiSpec == 'null:') return loc;
            //this.debug('original: ' + loc.spec);
            var asciiPath = loc.asciiSpec.substring(loc.prePath.length);
            var r = loc.scheme + '://';
            if (loc.userPass) r += loc.userPass + '@';
            r += loc.host;
            var skip = 0;
            for each (var c in asciiPath) {
                if (skip > 0) {
                    skip--;
                    r += c;
                    continue;
                }

                switch (c) {
                case 'a': case 'b': case 'c': case 'd': case 'e':
                case 'f': case 'g': case 'h': case 'i': case 'j':
                case 'k': case 'l': case 'm': case 'n': case 'o':
                case 'p': case 'q': case 'r': case 's': case 't':
                case 'u': case 'v': case '0': case '1': case '2':
                case '3': case '4': case '5': case '6': case '7':
                case '8': case '9': 
                    r += c;
                    break;

                case 'w': case 'x': case 'y': case 'z':
                    r += 'w' + c.toLowerCase();
                    break;

                case 'A': case 'B': case 'C': case 'D': case 'E':
                case 'F': case 'G': case 'H': case 'I': case 'J':
                case 'K': case 'L': case 'M': case 'N': case 'O':
                case 'P': case 'Q': case 'R': case 'S': case 'T':
                case 'U': case 'V': case 'W': case 'X': case 'Y': case 'Z':
                    r += 'x' + c.toLowerCase();
                    break;

                case '/':
                    r += '.';
                    break;

                case '%':
                    r += 'y';
                    skip = 2;
                    break;

                default:
                    r += 'z' + c.charCodeAt(0).toString(16);
                    break;
                }
            }
            r += '.illegal';
            if (loc.port > -1)
                r += ':' + loc.port;
            r += '/';
            //this.debug('transformed: ' + r);
            return this.makeURI(r);
        },

        getLanguageName : function(s) {
            try {
                return this.oLanguageBundle.GetStringFromName(s);
            } catch (e) {
                return "Unknown";
            }
        },

        getRegionName : function(s) {
            try {
                return this.oRegionBundle.GetStringFromName(s);
            } catch (e) {
                return "Unknown";
            }
        },

        nameLanguageCode : function(s) {
            var parts = s.trim().toLowerCase().split(/[-_.]/);
            //this.debug('code ' + s + ' has ' + parts.length + ' parts');
            // TODO: Localize this string.
            if (parts.length == 0) return "Unknown";
            //this.debug("looking for lang " + parts[0]);
            var r = "" + this.getLanguageName(parts[0]);
            if (parts.length == 1)
                return r;
            //this.debug("looking for region " + parts[1]);
            r += " (" + this.getRegionName(parts[1]) + ")";
            return r;
        },

        TranslateString: function(sName, sVar1, sVar2, sVar3)
        {
            var sResult = "";
            if(this.oStringBundle)
            {
                sResult  = this.oStringBundle.GetStringFromName(sName);
                if (sVar1 || sVar1 == 0)
                    sResult = sResult.replace(/%1/g, sVar1);
                if (sVar2 || sVar2 == 0)
                    sResult = sResult.replace(/%2/g, sVar2);
                if (sVar3 || sVar3 == 0)
                    sResult = sResult.replace(/%3/g, sVar3);
            }
            return sResult;
        },

        getEnvVar : function(name) {
            return this.Components.classes["@mozilla.org/process/environment;1"]
                .getService(this.Components.interfaces.nsIEnvironment)
                .get(name);
        },

        setEnvVar : function(name, value) {
            return this.Components.classes["@mozilla.org/process/environment;1"]
                .getService(this.Components.interfaces.nsIEnvironment)
                .set(name, value);
        },

        getExtensionFolder : function() {
            return this.Components.classes["@mozilla.org/extensions/manager;1"]
                .getService(this.Components.interfaces.nsIExtensionManager)
                .getInstallLocation(this.extId)
                .getItemLocation(this.extId);
        },

        msgBox : function(title, text) {
            this.Components.classes["@mozilla.org/embedcomp/prompt-service;1"]
                .getService(this.Components.interfaces.nsIPromptService)
                .alert(window, title, text);
        },

        loadScript : function(url) {
            var jsLoader =
                this.Components.classes["@mozilla.org/moz/jssubscript-loader;1"]
                .getService(this.Components.interfaces.mozIJSSubScriptLoader);
            jsLoader.loadSubScript(url);
        },

        isVisitied : function(oURI) {
            oGlobalHistory =
                this.Components.classes["@mozilla.org/browser/global-history;2"]
                .getService(this.Components.interfaces.nsIGlobalHistory2);
            return oURI ? oGlobalHistory.isVisited(oURI) : false;
        },

        // Get the accept_languages preferences as an array of items,
        // since this might be a localized pref we need to try and get
        // it as a complex first.
        GetAcceptLanguages: function()
        {
            var oALPref =
                this.Components.classes["@mozilla.org/preferences-service;1"]
                .getService(this.Components.interfaces.nsIPrefService)
                .getBranch("intl.");
            var aResult = new Array();
            var sResult = "";
            try {
                sResult =
                    oALPref.getComplexValue("accept_languages",
                                            this.Components.interfaces
                                            .nsIPrefLocalizedString).data;
            } catch (e) {
                sResult = "";
            }
            sResult = oALPref.getCharPref("accept_languages");
            if (sResult != "") aResult = sResult.split(",");
            return aResult;
        },

        // Saves 'content' to file 'filepath'. Note that filepath
        // needs to be a real file path, not a chrome path.
        saveToFile : function(filepath, content) {
            var file = this.getFile(filepath);

            if (!file.exists()) {
                file.create(this.Components.interfaces
                            .nsIFile.NORMAL_FILE_TYPE, 420);
            }
            var outputStream = this.Components.classes
            ["@mozilla.org/network/file-output-stream;1"]
                .createInstance(this.Components.interfaces.nsIFileOutputStream);

            outputStream.init( file, 0x04 | 0x08 | 0x20, 420, 0 );
            var result = outputStream.write( content, content.length );
            outputStream.close();
        },

        // Returns the content of 'file' as string. The 'file'
        // parameter should be a nsIFile object.
        getFileContents : function(file) {
            var fstream = this.Components.classes
            ["@mozilla.org/network/file-input-stream;1"]
                .createInstance(this.Components.interfaces.nsIFileInputStream);
            var sstream =
                this.Components.classes["@mozilla.org/scriptableinputstream;1"]
                .createInstance(this.Components.interfaces
                                .nsIScriptableInputStream);
            fstream.init(file, -1, 0, 0);
            sstream.init(fstream);

            var data = '';

            var str = sstream.read(4096);
            while (str.length > 0) {
                data += str;
                str = sstream.read(4096);
            }
            fstream.close();
            return data;
        },

        // Returns the content of 'file' as string. The 'file'
        // parameter should be a nsIFile object.
        getBinaryFileContents : function(file) {
            var fstream = this.Components.classes
            ["@mozilla.org/network/file-input-stream;1"]
                .createInstance(this.Components.interfaces.nsIFileInputStream);
            fstream.init(file, -1, -1, false);

            var bstream = this.Components.classes
            ["@mozilla.org/binaryinputstream;1"]
                .createInstance(this.Components.interfaces
                                .nsIBinaryInputStream);
            bstream.setInputStream(fstream);

            var bytes = bstream.readBytes(bstream.available());
            fstream.close();
            return bytes;
        },

        startProcess: function(filename, args) {
            var file = this.getFile(filename);

            args = args ? args : [];

            if (file.exists()) {
                var nsIProcess =
                    this.Components.classes["@mozilla.org/process/util;1"]
                    .getService(this.Components.interfaces.nsIProcess);
                nsIProcess.init(file);
                nsIProcess.run(false, args, args.length);
            } else {
                throw Error(this.stringSubst("File '%1' does not exist!",
                                             filename));
            }
        },

        // Simulates a double click on the file in question.
        launchFile : function(filepath) {
            var f = this.getFile(filepath);
            f.launch();
        },

        // Gets a local file reference, return the interface
        // nsILocalFile.
        getFile : function(filepath) {
            var f = this.Components.classes["@mozilla.org/file/local;1"]
                .createInstance(this.Components.interfaces.nsILocalFile);
            f.initWithPath(filepath);
            return f;
        },

        filePicker : function(name) {
            var nsIFilePicker = this.Components.interfaces.nsIFilePicker;
            var FilePicker =
                this.Components.classes["@mozilla.org/filepicker;1"]
                .createInstance(nsIFilePicker);
            FilePicker.init(window, name, nsIFilePicker.modeOpen);
            FilePicker.appendFilters(nsIFilePicker.filterXML);
            var ReturnFP = FilePicker.show();
            if (ReturnFP == nsIFilePicker.returnOK ||
                ReturnFP == nsIFilePicker.returnReplace) {
                var Picker = Array();
                Picker['file'] = FilePicker.file;
                Picker['path'] = FilePicker.file.path;
                Picker['fileURI'] = this.ioService.newFileURI(Picker['file']);
                return Picker;
            } else {
                return false;
            }
        },

        // Returns all elements that match the query sent in. The root
        // used in the query is the window.content.document, so this
        // will only work for html content.
//         xpath : function(query, window) {
//             return window.content.document.evaluate
//             (query, window.content.document, null,
//              XPathResult.UNORDERED_NODE_SNAPSHOT_TYPE, null);
//         },

        // Adjust the set of classes on an element.
        setClassScale : function(sElemPath, sAttrName, sAttrNS, fNSresolver,
                                 sBase, iMin, iMax, iVal, doc, window) {
            var elem = doc.evaluate(sElemPath, doc, fNSresolver,
                                    window.XPathResult.FIRST_ORDERED_NODE_TYPE,
                                    null);
            if (!elem) return;
            elem = elem.singleNodeValue;
            if (!elem) return;
            var attrVal = sAttrNS ?
                elem.getAttributeNS(sAttrNS, sAttrName) :
                elem.getAttribute(sAttrName);

            // Get the current set of classes.
            var classes = { };
            if (attrVal) {
                for each (var s in attrVal.trim().split(/\s+/)) {
                    classes[s] = true;
                }
            }

            // Adjust the set.
            for (var i = iMin; i < iVal; i++) {
                delete classes[sBase + "-under-" + i];
                classes[sBase + "-over-" + i] = true;
            }
            delete classes[sBase + "-under-" + iVal];
            delete classes[sBase + "-over-" + iVal];
            for (var i = iVal + 1; i <= iMax; i++) {
                classes[sBase + "-under-" + i] = true;
                delete classes[sBase + "-over-" + i];
            }

            // Write the classes back to the element.
            attrVal = "";
            for (var i in classes) {
                if (classes[i])
                    attrVal += " " + i;
            }
            if (sAttrNS)
                elem.setAttributeNS(sAttrNS, sAttrName, attrVal);
            else
                elem.setAttribute(sAttrName, attrVal);
        },

        setStatusBarText : function(obj, window) {
            var text;
            if (obj.getAttribute) { // This is an element.
                text = obj.getAttribute('statustext');
            } else {
                text = obj;
            }

            if (text) {
                var statusbar = window.top.document
                    .getElementById('statusbar-display');

                if (statusbar) {
                    statusbar.oldtext = statusbar.getAttribute('label');
                    statusbar.setAttribute('label', text);
                }
            }
        },

        clearStatusBarText : function(window) {
            var statusbar =
                window.top.document.getElementById('statusbar-display');

            if (statusbar) {
                statusbar.setAttribute('label', statusbar.oldtext ?
                                       statusbar.oldtext : '');
            }
        },

        getStatusBarText : function(window) {
            var statusbar =
                window.top.document.getElementById('statusbar-display');

            if (statusbar) {
                return statusbar.getAttribute('label');
            }
            return '';
        },
    
        GetString: function(sPropertyName, document)
        {
            var oStrings = document.getElementById("standardsitemap-strings");
            if (oStrings) return oStrings.getString(sPropertyName);
            return "Error unknown property: " + sPropertyName;
        },
    
        GetAppVersion: function()
        {
            return this.Components.classes["@mozilla.org/xre/app-info;1"]
                .getService(this.Components.interfaces.nsIXULAppInfo).version;
        },
    
        // Generates a probably unique id in a 8 long HEX format
        // (09AFFA90) (there is a chance that it might not be unique I
        // guess...)
        GenerateID: function()
        {
            return ((((1+Math.random())*0x10000)|0).toString(16).substring(1) +
                    (((1+Math.random())*0x10000)|0).toString(16).substring(1))
                .toUpperCase();
        },
    
        ReadCSS: function(window, sURI)
        {
            try
            {
                var oRequest = new window.XMLHttpRequest();
                oRequest.open("GET", sURI, false);
                oRequest.overrideMimeType("text/css");
                oRequest.send(null);

                if (oRequest.status != 404)
                {
                    if (oRequest.responseText) return oRequest.responseText;
                }
            } catch (e) {
                this.debug('Obtaining CSS from ' + sURI + ': ' + e);
                return "";
            }
            return "";
        },
    
        IsLastWindow: function()
        {
            var oWW = this.Components
                .classes["@mozilla.org/embedcomp/window-watcher;1"]
                .getService(this.Components.interfaces.nsIWindowWatcher);
            var oWEnum = oWW.getWindowEnumerator();
            var iCount = 0;
            var bResult = true;
            while (oWEnum.hasMoreElements)
            {
                iCount ++;
                oWEnum.getNext();
                if (iCount > 10) // Fail save...
                {
                    bResult = false;
                    break;
                }
            }
            if (iCount <= 1) bResult = true;
            return bResult;
        },

        // Returns true if the string contains the substring str.
        stringContains : function (s, str) {
            return s.indexOf(str) != -1;
        },

        stringTrim : function(s) {
            return s.replace(/^\s*|\s*$/gi, '');
        },

        stringStartsWith : function(b, s) {
            return (b.length >= s.length && b.substr(0, s.length) == s);
        },

        stringEndsWith : function(b, s) {
            return (b.length >= s.length && b.substr(b.length - s.length) == s);
        },

        // Inserts the arguments into the string. E.g. if the string
        // is "Hello %1" then stringSubst("Hello %1", 'johnny') would
        // return "Hello johnny"
        stringSubst : function() {
            var s = arguments[0];
            for (var i = 1; i < arguments.length; i++) {
                var nr = "%" + i;
                var repl;
                if (arguments[i] == null) {
                    repl = "null";
                } else if (arguments[i] == undefined) {
                    repl = "undefined";
                } else {
                    repl = arguments[i];
                }
                s = s.replace(new RegExp(nr, "g"), repl);
            }
            return s;
        },

        xulHide : function(b) {
            b.style.display = 'none';
        },

        xulShow : function(b) { 
            b.style.display = '';
        },

        xulEnable : function(b) {
            b.setAttribute('disabled', false);
        },

        xulDisable : function(b) {
            b.setAttribute('disabled', true);
        },

        xulAddClass : function(b, Class) {
            var current = b.getAttribute('class') + '';
            var classes = current.split(' ');
            for (var i = 0; i < classes.length; i++)
                if (classes[i] == Class)
                    return;
            b.setAttribute('class', current + ' ' + Class);
        },

        xulRemoveClass : function(b, Class) {
            var current = b.getAttribute('class') + '';
            var classes = current.split(' ');
            var newclass = '';
            for each (var name in classes) {
                if (name != Class) {
                    if (newclass != '')
                        newclass += ' ';
                    newclass += name;
                }
            }
            b.setAttribute('class', newclass);
        },

        xulAddProperty : function(b, Property) {
            var current = b.getAttribute('properties') + '';
            var properties = current.split(' ');
            for each (var p in properties) {
                if (p == Property)
                    return;
            }
            b.setAttribute('properties', current + ' ' + Property);
        },

        xulRemoveProperty : function(b, Property) {
            var current = b.getAttribute('properties') + '';
            var properties = current.split(' ');
            var newproperties = '';
            for each (var name in properties) {
                if (name != Property) {
                    if (newproperties != '')
                        newproperties += ' ';
                    newproperties += name;
                }
            }
            b.setAttribute('properties', newproperties);
        },

        xulRemoveAllChildren : function(b) {
            while (b.firstChild) {
                b.removeChild(b.firstChild);
            }
        },


        xmlGetElementById : function (b, myID) {
            return b.evaluate('//*[@xml:id="'+myID+'"]', b,
                              function () { return this.xmlNS; },
                              XPathResult.FIRST_ORDERED_NODE_TYPE,
                              null).singleNodeValue;
        },

        Map : function(size, fHash, fEquals) {
            if (typeof size != "integer")
                size = 271;
            if (typeof fHash != "function")
                fHash = function(obj) {
                    var r = 0;
                    var s = toString(obj);
                    var len = s.length;
                    for (var i = 0; i < len; i++)
                        r += s.charCodeAt(i);
                    return r;
                };
            if (typeof fEquals != "function")
                fEquals = function(obj1, obj2) {
                    return obj1 == obj2;
                };
            this.fHash = fHash;
            this.fEquals = fEquals;
            this.size = size;
            this.clear();
        },

        // TODO: Does JavaScript have any sort of general Map class,
        // i.e. mapping object to object?  Seems that an associative
        // array can only handle strings, i.e. it converts its keys to
        // strings beforehand.  If there is a native Map class, use
        // that instead of this.
        newMap : function(size, fHash, fEquals) {
            return new (this.Map)(size, fHash, fEquals);
        },
    });


    Application.storage.get("org.standard-sitemap.library", null)
        .Map.prototype = {
            clear : function() {
                this.aBuckets = [];
            },

            bucket : function(obj) {
                var i = this.fHash(obj) % this.size;
                if (!this.aBuckets[i])
                    this.aBuckets[i] = [];
                return this.aBuckets[i];
            },

            foreach : function(fAction) {
                for each (var bucket in this.aBuckets) {
                    for each (var entry in bucket) {
                        fAction(entry.key, entry.value);
                    }
                }
            },

            contains : function(key) {
                return this.findInBucket(this.bucket(key), key);
            },

            findInBucket : function(bucket, key) {
                for each (var entry in bucket) {
                    if (this.fEquals(key, entry.key))
                        return entry;
                }
                return null;
            },

            findPositionInBucket : function(bucket, key) {
                for (var i in bucket) {
                    var entry = bucket[i];
                    if (this.fEquals(key, entry.key))
                        return i;
                }
                return -1;
            },

            get : function(key) {
                var bucket = this.bucket(key);
                var entry = this.findInBucket(bucket, key);
                if (entry) return entry.value;
            },

            // Encapsulate the mapped value in an object with one
            // field 'value', or return null.  This allows the caller
            // to distinguish between the key being missing and the
            // key mapping explicitly to null.
            getNull : function(key) {
                var bucket = this.bucket(key);
                var entry = this.findInBucket(bucket, key);
                if (entry)
                    return { value : entry.value };
            },

            put : function(key, value) {
                if (typeof value == "undefined")
                    return this.remove(key);
                var bucket = this.bucket(key);
                var entry = this.findInBucket(bucket, key);
                if (entry) {
                    var old = entry.value;
                    entry.value = value;
                    return old;
                }

                entry = { key : key, value : value };
                bucket.push(entry);
            },

            remove : function(key) {
                var bucket = this.bucket(key);
                var i = this.findPositionInBucket(bucket, key);
                if (i < 0)
                    return;
                var entry = bucket[i];
                bucket.splice(i, 1);
                return entry.value;
            }
        };

    Application.storage.get("org.standard-sitemap.library", null)
        .initialize();
    Application.storage.get("org.standard-sitemap.library", null)
        .debug('Library installed');
}
